/*********************************************************************

SIMWORLD
Version 4.0 Beta

Basic Rules

Copyright (C) 2002 Matthias Scheutz (mscheutz@nd.edu)
Artificial Intelligence and Robotics Laboratory
Department of Computer Science and Engineering 
University of Notre Dame, USA
http://www.nd.edu/~mscheutz/ 
http://www.nd.edu/~airolab/ 
All rights reserved.

This program is licensed and distributed under the terms stated
in the accompanying COPYRIGHT file.

Last modified: 11-27-02

*********************************************************************/
compile_mode :pop11 +strict;

;;; this is for GUI mode
global vars 
    printdatabase,

    ;;; this is from simworld.p
    sim_pause_demo,

;

;;; these two rulesets are only for debugging...
define :ruleset printagentdatabase;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  ;;; if food is there and you are not eating yet,
  ;;; stop and start eating
  RULE printingdb
        ==>
    [POP11 if (sim_pause_demo and printdatabase(sim_myself)) then
               prb_print_table(sim_data(sim_myself));
           endif;]
enddefine;

;;;-------------------------------------------------------------------------
;;; the BASIC modules

;;; the EAT module
define :ruleset eating_ruleset;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  ;;; if food is there and you are not eating yet,
  ;;; stop and start eating
  RULE start_eating
    [ingest food] [NOT halting_for eating]
        ==>
    [do halting_for eating]
    [do eating]
    [STOP]

  ;;; if no food is present and you are still eating, move on
  RULE end_eating
    [NOT ingest food] [halting_for eating]
        ==>
    [NOT halting_for eating]
    [STOP]

  ;;; if food is there and you are eating, continue eating
  RULE continue_eating
    [ingest food] [halting_for eating]
        ==>
    [do eating]
enddefine;

;;; the DRINK module
define :ruleset drinking_ruleset;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  ;;; if water is there and you are not drinking yet,
  ;;; stop and start drinking
  RULE start_drinking
    [ingest water] [NOT halting_for drinking]
        ==>
    [do halting_for drinking]
    [do drinking]
    [STOP]

  ;;; if no water is present and you are still drinking, move on
  RULE end_drinking
    [NOT ingest water] [halting_for drinking]
        ==>
    [NOT halting_for drinking]
    [STOP]

  ;;; if water is there and you are drinking, continue drinking
  RULE continue_drinking
    [ingest water] [halting_for drinking]
        ==>
    [do drinking]
enddefine;

;;; rules for the MOVE module
define :ruleset moving_ruleset;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  ;;; if sensory space is not already mapped to motor space, do it now
  ;;; this rule will always be executed
  RULE compute_motor_scheme
    [NOT nomapping]
        ==>
    [do map_motor]

  ;;; two speed rules to have different speeds when not in emergency state
  ;;; normal speed
  RULE determine_speed
    [moving] [speed ?x] [WHERE x < 4]
    [NOT alarm =]
        ==>
    [do speeding_up]

  ;;; speed when alarm is set
  RULE determine_speed
    [moving] [speed ?x] [WHERE x < 1]
        ==>
    [do speeding_up]

  ;;; if you are not halting for whatever reason
  ;;; and you should move, move!
  ;;; always speed up till you reach maximum speed if you are moving
  RULE get_moving
    ;;; check if in start state and halt is true
    [moving] [NOT halting_for =]
        ==>
    [do moving]
    [STOP]

  ;;; if you are not halting (for whatever reason), try to move
  RULE try_to_move
    [NOT moving] [NOT halting_for =]
        ==>
    [moving]
enddefine;


;;; rules for the WATER ALARM module
;;; the critical water level is 200
define :ruleset coolantalarm_ruleset;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  RULE initiate_water_alarm
    ;;; check if in start state and halt is true
    [NOT alarm water] [NOT alarm energy] [waterlevel ?x]
    [WHERE x < 400]
        ==>
    [alarm water]
    [nomapping]
    [do slowing_down]
    [do slowing_down]
    [do slowing_down]
    [STOP]

  RULE check_for_water_alarm
    ;;; check if in start state and halt is true
    [alarm water] [waterlevel ?x]
    [WHERE x > 400]
        ==>
    [NOT alarm water]
    [NOT nomapping]
    [STOP]

  RULE focus_on_water
    [alarm water] [NOT ingest water]
       ==>
    ;;; do it directly for speed reasons
    [POP11 watervector(sim_myself)(1) -> movevector(sim_myself)(1);
           watervector(sim_myself)(2) -> movevector(sim_myself)(2);]
    [STOP]
enddefine;

;;; since this ALARM is executed after the water alarm, it has precedence
;;; rules for the ENERGY ALARM module
;;; critical energy level = 300
define :ruleset energyalarm_ruleset;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  RULE initiate_energy_alarm
    ;;; check if in start state and halt is true
    [NOT alarm energy] [energylevel ?x]
    [WHERE x < 600]
        ==>
    [alarm energy]
    [nomapping]
    [do slowing_down]
    [do slowing_down]
    [do slowing_down]
    [STOP]

  RULE check_for_energy_alarm
    ;;; check if in start state and halt is true
    [alarm energy] [energylevel ?x]
    [WHERE x > 600]
        ==>
    [NOT nomapping]
    [NOT alarm energy]
    [STOP]

  ;;; remove any other vector and leave only food
  ;;; also take priority over alarm mechanism
  RULE focus_on_energy
    [alarm energy] [NOT ingest food]
       ==>
    [NOT alarm water]    ;;; suppress coolant alarm as energy is more vital
    [POP11 foodvector(sim_myself)(1) -> movevector(sim_myself)(1);
           foodvector(sim_myself)(2) -> movevector(sim_myself)(2);]
    [STOP]
enddefine;

;;; TODO: implement the reflex also as a (very strong) FORCE!!!
;;; the COLLIDE module
define :ruleset collidereflex_ruleset;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  ;;; always update the reflex activation...
  RULE start_reflex1
    [collide start] [agentcollision %true%] [ignoreagent %false%]
        ==>
    [halting_for reflex]
    [NOT collide start] [collide reflex]
    [POP11 true -> agent_alarm(sim_myself)]
    [STOP]

  ;;; if ignoreagent is true, then go straight to moving 0 to skip the
  ;;; reflex itself
  RULE start_reflex2
    [collide start] [agentcollision %true%] [ignoreagent %true%]
        ==>
    [halting_for reflex]
    [NOT collide start] [collide move] [moving 0]
    [POP11 true -> agent_alarm(sim_myself)]
    [STOP]

  RULE start_reflex3
    [collide start] [obstcollision %true%]
        ==>
    [halting_for reflex]
    [NOT collide start] [collide reflex]
    [POP11 true -> obst_alarm(sim_myself)]
    [STOP]

  RULE reflex_reflex
    [collide reflex] [collision %true%]
        ==>
    [do newrandomheading 30 165]
    [STOP]

  RULE reflex_move
    [collide reflex] [collision %false%]
        ==>
    [NOT collide reflex] [collide move]
  ;;;  [moving <put in the time steps here>]
    [moving 1]
    [do reflex_basic_agent]
    [STOP]

  RULE move_start
    [collide move] [moving 0]
       ==>
    [NOT moving 0]
    [NOT collide move] [collide start]
    [NOT halting_for reflex]
    [STOP]

  RULE move_move
    [collide move] [moving 1]
        ==>
    [NOT moving 1] [moving 0]
    [do reflex_basic_agent]
    [STOP]

  RULE startup_collide
    [NOT collide =]
        ==>
    [collide start]
enddefine;

;;; the cleanup rules, empties the database of all volatile information
define :ruleset cleanup_ruleset;
    [DLOCAL [prb_allrules = true] [prb_sortrules = false]];

  RULE cleanup
    ==>
  [POP11 lvars ag = sim_data(sim_myself);
         sim_flush_data([collision ==],ag);
         sim_flush_data([obstcollision ==],ag);
	 sim_flush_data([ignoreagent ==],ag);
	 sim_flush_data([agentdiff ==],ag);
	 sim_flush_data([observe ==],ag);
         sim_flush_data([agentcollision ==],ag);
         sim_flush_data([othercollision ==],ag);
         sim_flush_data([sameagentcollision ==],ag);
         sim_flush_data([agentvector ==],ag);
         sim_flush_data([othervector ==],ag);
         sim_flush_data([samevector ==],ag);
         sim_flush_data([obstaclevector ==],ag);
         sim_flush_data([foodvector ==],ag);
         sim_flush_data([watervector ==],ag);
         sim_flush_data([touch ==],ag);
         sim_flush_data([ingest ==],ag);
         sim_flush_data([obstacle ==],ag);
         sim_flush_data([same ==],ag);
         sim_flush_data([other ==],ag);
         sim_flush_data([sound ==],ag);
         sim_flush_data([vision ==],ag);
         sim_flush_data([food ==],ag);
         sim_flush_data([water ==],ag);
         sim_flush_data([energylevel ==],ag);
         sim_flush_data([waterlevel ==],ag);
         sim_flush_data([heading ==],ag);
         sim_flush_data([speed ==],ag);
  ]
enddefine;
