/*********************************************************************

SIMWORLD                                                */ global vars
Version = '4.0 Beta'                                               ;/*

Simworld GUI

Copyright (C) 2002 Matthias Scheutz (mscheutz@nd.edu)
Artificial Intelligence and Robotics Laboratory
Department of Computer Science and Engineering
University of Notre Dame, USA
http://www.nd.edu/~mscheutz/
http://www.nd.edu/~airolab/
All rights reserved.

This program is licensed and distributed under the terms stated
in the accompanying COPYRIGHT file.

Last modified: 11-27-02

*********************************************************************/

;;; this may have to be adjusted
constant path_to_simworld_files = sys_fname_path(popfilename);

;;; compile the basic simulator
compile(path_to_simworld_files dir_>< 'simworld.p');

;;; ------------------- External libraries --------------------------
uses sim_picagent;
uses rc_graphplot;

;;; --------------- GLOBAL SIMULATION PARAMETERS -------------------
;;; define the server host and its port as default
;;; start in client mode, and assume that server and port are
;;; set correctly by the calling program if different from the
;;; default values below

global vars
    ;;; various graphical aides can be turned on or off
    SHOWSTATE = false, ;;; display food and enery levels
    SHOWVISION = false,   ;;; indicates whether vision range should be drawn
    SHOWSONAR = false,    ;;; indicates whether sonar range should be drawn
    SHOWSMELL = false,    ;;; indicates whether smell range should be drawn
    SHOWVECTORS = false,  ;;; draws the respective vectors
    SHOWPATH  = false,	  ;;; shows the paths of agents

    ;;; the simulation windows
    sim_windowlist = [],

    ;;; window location on screen
    window_x = "left",
    ;;; window_y = "top",
    window_y = 40,
    world_window_x = "right",
    ;;; world_window_y = "top",
    world_window_y = 40,

    ;;; half width and height of window
    window_lim_x = 300,
    window_lim_y = 300,
    world_window_lim_x = 300,
    world_window_lim_y = 300,
	
    ;;; specify mapping between simulation world and picture world
    sim_picxorigin = 0,
    sim_picyorigin = 0,

    ;;; the scale of the graphical representation
    sim_picxscale = world_window_lim_x/world_lim_x,
    sim_picyscale = world_window_lim_y/world_lim_y,

    ;;; window size, adjusted by the scale variables
    window_width = window_lim_x*2,
    window_height = window_lim_y*2,
    world_window_width = world_window_lim_x*2,
    world_window_height = world_window_lim_y*2,

    ;;;             rc_xorigin,      rc_yorigin, rc_xscale, rc_yscale
    window_frame = {%window_lim_x, window_lim_y, 1, -1 %},
    world_window_frame = {%world_window_lim_x,world_window_lim_y,sim_picxscale,-sim_picyscale %},

    ;;; This four element vector is used for the sim_picframe slot
    ;;; of the main simulation window. See HELP SIM_PICAGENT
    window_transform = {^sim_picxorigin ^sim_picyorigin 1 1},

	world_window_transform =
        {0 0 ^sim_picxscale ^sim_picyscale},

	world_window_title = 'WorldView',

    window_title = 'SimWorld ' >< Version >< ' -- (c) by Matthias Scheutz',

    ;;; if this is made true the simulation will pause at the end of
    ;;; the current time-slice
    sim_pause_demo = false,
	sim_delay = 0,
    sim_tracking_list = newproperty([], 16, [], "perm"),
    lastobject = undef,   ;;; used in the command interpreter to keep track of last object referred to
;

;;; ----------------- Load the other simworld files -----------------------
;;; this loads th agents' GUI files
;;; compile('/afs/nd.edu/user40/airolab/poplog/local/simworld/src/basicagentsGUI.p');
;;; compilehere basicagentsGUI.p ;
compile(path_to_simworld_files dir_>< 'basicagentsGUI.p') ;

;;;--------- override various SIMAGENT tracing functions --------------

define :method sim_agent_running_trace(agent:sim_agent);
    ;;; This is applied to each sim_agent (but not object) before running its
    ;;; rules, and after incoming messages and sensor data have beeb added
    ;;; to the agent's database.
    ;;; Default is to do nothing unless the variable demo_data_trace
    ;;; is made true, in which case the whole database is printed out.
    ;;; Redefine for individual subclasses to make printing more selective.
enddefine;

define :method sim_agent_messages_out_trace(agent:sim_agent);
    ;;; At the end of the time slice show the messages waiting to
    ;;; be delivered from this agent to other agents.
enddefine;

define :method sim_agent_messages_in_trace(agent:sim_agent);
    ;;; If demo_message_trace is true, show the incoming messages waiting
    ;;; to be processed just before the agent runs its rulesystem
enddefine;

define :method sim_agent_actions_out_trace(object:sim_object);
    ;;; By default, no tracing for static objects
enddefine;

define :method sim_agent_actions_out_trace(agent:sim_agent);
    ;;; Called at the end of each agent's time slice, to show which
    ;;; actions have been transferred from the agent's database to its
    ;;; sim_actions slot, so that they can be run at the end of the
    ;;; time slice.
enddefine;

define :method sim_agent_action_trace(agent:sim_agent);
enddefine;

define :method sim_agent_rulefamily_trace(object:sim_object, rulefamily);
    ;;; do nothing
enddefine;

define :method sim_agent_rulefamily_trace(agent:sim_agent, rulefamily);
    ;;; In each time slice, for each agent or object, before each
    ;;; ruleset or rulefamily is run, print out the ruleset or
    ;;; rulefamily contents, and also the contents of the database,
    ;;; but only if demo_startrules_trace is true.
enddefine;

define :method sim_agent_endrun_trace(agent:sim_object);
    ;;; After an object or agent has had its rulesystem run
    ;;; this can be used to print out information such as its
    ;;; current database contents.
enddefine;

define sim_scheduler_finished(objects, cycle);
enddefine;

;;; overwrites the same function in simworld.p
;;; implements a primitive command interpreter here to manipulate and inspect objects
define sim_scheduler_pausing_trace(objects, cycle);
    ;;; Invoked by sim_scheduler at the end of each time slice.
    ;;; It allows arbitrary tracing or pausing mechanisms defined by the
    ;;; user to run, e.g. to interrogate the internal databases of
    ;;; selected objects, or to save the state of the pictures on the
    ;;; screen, etc.

    ;;; first finish up various book keeping things for this cycle
    finish_up_cycle(objects,cycle);

    ;;; NEW STUFF IN THIS FUNCTION
    ;;; "l" (for last) in a comand always refers to the last one used

    if sim_pause_demo then
            lvars cmdln,item,agent,subitem;
            appproperty(sim_tracking_list, procedure(name,vals);
                    npr('Object ' >< name); lvars agent;
                    for agent in objects do
                        if sim_name(agent) == name then
                            unless (sim_status(agent) == "dead") then
                                applist(vals,procedure(x);
                                          npr('Slot ' >< x >< ': ' >< valof(x)(agent));
                                          endprocedure);
                            else
                                npr('Removed dead object ' >< name);
                                [] -> sim_tracking_list(name);
                            endunless;
                            quitloop(1);
                        endif;
                    endfor;
                    pr(newline);endprocedure);
            ;;; parse the command line
            pr(cycle >< ' ');
            while not(null(readline()->>cmdln)) do
              lvars cmd = front(cmdln);
              if cmd = "help" then
		  npr('Available commands are:');
                  npr('  make <object> - creates a new object of type "object"');
		  npr('  see <object> <slot> - print the content of the object\'s slot');
		  npr('  set <object> <slot> <value> - set the slot of the object to value');
		  npr('  move <object> <x> <y> - moves the object to the (x,y) position');
		  npr('  del <object> - removes the object from the simulation');
		  npr('  pdb <agent> - print the agent\'s database');
                  npr('  adb <agent> <item> - adds the item to the agent\'s database');
                  npr('  ddb <agent> <item> - deletes the item from the agent\'s database');
		  npr('  toff - turns tracking off for all objects');
		  npr('  toff <object> - turns tracking off for all slots of the object');
		  npr('  ton <object> <slot1> <slot2> ... - turns tracking on for the respective slots of the object');
                  npr('  pop11 <string command> - compiles the POP11 command within single quotes');
		  npr('---------------------------------------------------------------------------------------------');
              elseif cmd == "toff" and null(tl(cmdln)) then
                  clearproperty(sim_tracking_list);
                  npr('Tracking is turned off (all slots are removed from tracking list)');
              elseif null(tl(cmdln)) then
                  npr('Need parameters for cmd ' >< cmd >< ' or illegal command');
              else
                lvars name = front(back(cmdln)), params = back(back(cmdln)),
                     found = false;
                if cmd == "pop11" then
                    if isstring(name) then
                        compile(stringin(name));
                    else
                        npr('Argument must be string');
                    endif;
                elseif cmd == "save" then
		    npr('Cannot save simulation state in graphics mode!');
              	elseif cmd == "load" then
                     npr('Cannot load simulation state in graphics mode!');
		elseif cmd == "make" then
                    lvars GUIconstruct = constructors(name);
                    if isprocedure(GUIconstruct) then
                        conspair(newobject(constructors,name),
                                 sim_object_add_list) -> sim_object_add_list;
                    else
						npr('No object type ' >< name >< ' is defined for SimWorld');
                    endif;
                else
                  for agent in objects do
				 	if (name == "l" and lastobject == sim_name(agent)) or
                       (sim_name(agent) == name) then
					  ;;; check whether a name was used to refer to an agent
                      if name /== "l" then
                          name -> lastobject;
                      else
                          lastobject -> name;
                      endif;
                      true -> found;
                      ;;; check for all possible commands
                      if cmd == "ton" then
                          if check_slots(params,agent) then  ;;;maybe check for already tracked slots
                              sim_tracking_list(name) <> params
                                  -> sim_tracking_list(name);
                              npr('Added slot(s) '>< params >< ' to tracking list');
                          endif;
                      elseif cmd == "toff" then
                          if null(params) then
                              [] -> sim_tracking_list(name);
                              npr('All slots of agent ' >< name >< ' are removed from tracking list');
                          else
                              delete(params(1),sim_tracking_list(name))->sim_tracking_list(name);
                              npr('Removed slot ' >< params(1) >< ' of agent ' >< name >< ' from tracking list');
                          endif;
                      elseif cmd == "see" then
                          ;;; show slot for agent
                          if null(params) then
                              npr('Which slot of agent ' >< name >< ' would you like to see:');
                              while null(readline()->>params) do
                                  npr('Which slot of agent ' >< name >< ' would you like to see:');
                              endwhile;
                          endif;
                          if check_slots(params(1),agent) then
                              npr('Slot ' >< params(1) >< ': ' >< valof(params(1))(agent));
                          endif;
                      elseif cmd == "set" then
                          ;;; set slot value for agent
                          if null(params) then
                            npr('Which slot of agent ' >< name >< ' would you like to set:');
                            while null(readline()->>params) do
                                npr('Which slot of agent ' >< name >< ' would you like to set:');
                            endwhile;
                          endif;
                          if check_slots(params(1),agent) then
                              if null(tl(params)) then
                                  npr('Need value for slot: ' >< params(1) >< ' - not set');
                                  quitloop(1);
                              endif;
                              if isword(params(2)) then
                                  valof(params(2)) -> (valof(params(1)))(agent);
                              else
                                  params(2) -> (valof(params(1)))(agent);
                              endif;
                              npr('Slot ' >< params(1) >< ': ' >< valof(params(1))(agent));
                          endif;
                      elseif cmd == "pdb" then
                          prb_print_table(sim_data(agent));
                      elseif cmd == "adb" then
                          prb_add_to_db(params,sim_data(agent));
                      elseif cmd == "ddb" then
                          dlocal prb_database = sim_data(agent);
                          prb_flush1(params);
		      elseif cmd == "move" then
                          (params(1),params(2)) -> sim_coords(agent);
		      elseif cmd == "del" then
                          if isbasic_agent(agent) then
                               dies_basic_agent(agent,'User intervention');
                          else
                               dies_entity(agent);
                          endif;
                      endif;
                      true -> found;
                      quitloop(1);
                    endif;
                  endfor;
                  unless found then
                      npr('Agent ' >< name >< ' does not exist');
                  endunless;
                endif;
              endif;
            endwhile;
        elseif isinteger(sim_delay) and sim_delay > 0 then
            syssleep(sim_delay)
        endif;
enddefine;


;;;------------------------------------------------------------------------------------------------------
/* This is completely new (no syntax check performed yet, should be added...)!

  The simulation now expects the following parameter format:

  [parameterlist entitylist resourcelist]

  The "parameterlist" contains basic information about the simulation, such as run time,
  file names, output formats, number of decimals printed, etc.,

    [[output {SCREEN | FILE <filename> {ASCII | HTML}]
     [runtime {<num> | <termproc}]
     [decimals <num>]
     [setup <setupproc>]
    ]

events fromcycle tocycle <proc>... -- to allow particular changes to happen

  The <termproc> is a procedure that takes as its argument the number of simulation cycles
  to decide whether the simulation should end.  It needs to set true -> sim_stopping_scheduler;
  <setupproc> is a function that gets passed a list of all initially created agents and can be
  used to do more simulation specific setup

  The "entitylist" is a list of entity records for each entity to be created at the
  beginning, which specifies what to do with it

  Each entry has the following format

    [<entity>
       [<keyword> <item1> <item2> ...]
       [<kewword> <item1> <item2> ...]
       ...
    ]

  where <entity> is a defined (derived) sim_object class;

  The "startup" keyword defines the kinds of entities that should be created at the beginning of
  the simulation.  The following forms are available:

    startup <num> &<initlist>
    startup [<initlist1> &<initproc1>] [<initlist2> &<initproc2>] ...

  <num> is the number of agents

  <initlist> is a list [[<slot1> <val1>] [<slot2> <val2>] ...] where <slot1>,<slot2>, etc. are
  slots in the class <entity> and <val1>,<val2>, etc. the corresponding values

  <initproc> is a procedure that will be called with the agent passed as an argument after the
  initalizing the slots values with initlist (if present)

  The "initial" keyword defines what should happend upon creation of a new agent of type <entity>:

    initial &<initproc>

  <initproc> will be passed either two arguments, the newly created
  agent and its parent, if procreation is asexual, or a list of its parents, if procreation
  is sexual.  This function can be used to perform "inheritance" on certain slot values, for example, or
  "crossover".  If "initial" is omitted, the agent will be created with its default values.

  The "transform" keyword defines whether agents of kind <entity> can turn into agents of kind <kind>.

    transform [<transprob> <kind1> &<transproc1>] [<transprob> <kind2> &<transproc2>] ...

  <transprob> is the probability that an agent will transform into type <kind>, which can be any class
  derived from sim_object.  The option transprob procedure takes as its two arguments the newly created
  agent of kind <kind> (which has only its default slots!!!), and the parent (of type <entity>).  When
  transformation takes place, no initialization occurs.

  The "mutate" keyword defines which slots should be mutated and how

    mutate [<mutprob1> [<slot1_1> <mutproc1_1>] [<slot1_2> <mutproc1_2>] ...]
           [<mutprob2> [<slot2_1> <mutproc2_1>] [<slot2_2> <mutproc2_2>] ...] ...

  where <mutprob> is the probability for mutation (upon creating of a new agent of type <entity>)
  of all the slots specified subsequenly by mutproc1, which is a procedure that takes two arguments,
  an agent and a slot.  Mutation always takes place after initialization and transformation.

  The "record" keyword is used to record statistical information about agents of type <entity>.  For every
  important event in an agent's life statistics can be recorded

    record [<freq1> [<slot1_1> &<recproc1_1>] [<slot2_1> &<recproc2_1>] ...]
           [<freq2> [<slot2_1> &<recproc2_1>] [<slot2_2> &<recproc2_2>] ...] ...

  The frequencies are for individual agents relative to their lifetime.  <freqs> are either numbers, or
  alternatively "birth" or "death".  The optional <recpro> returns a value to be recorded, otherwise the slot
  value is taken instead.  All records will be saved at then end of an agent's lifetime and
  can be turned into statistics classes at the end of the simulation.

  The "resourcelist" defines regular activities that should take place at the beginning and the end of
  simulation cycles.  It is a list of any of the following entries:

    [before <proc1> <proc2> ...
            [<freq1> [<kind1_1> &<initproc1_1>] [<kind1_2> &<initproc1_2>] ...]
            [<freq1> [<kind1_1> &<initproc1_1>] [<kind1_2> &<initproc1_2>] ...] ...

    [after  <proc1> <proc2> ...
            [<freq1> [<kind1_1> &<termproc1_1>] [<kind1_2> &<termproc1_2>] ...]
            [<freq1> [<kind1_1> &<termproc1_1>] [<kind1_2> &<termproc1_2>] ...] ...

  <proc> takes a list of all sim_objects and a cycle number, <initproc> takes an object to initalize,
  <termproc> takes an object to check for termination of the object

  these procedures are called

    [record [<freq1> [<kind1_1> <op1_1> <slot1_1_1> <slot1_1_2> ...] [<kind1_2> <op1_2> <slot1_2_1> <slot_1_2_2>] ...]
            [<freq2> [<kind2_1> <op2_1> <slot2_1_1> <slot2_1_2> ...] [<kind2_2> <op2_1> <slot2_2_1> <slot_2_2_2>] ...] ...

  <frequency> is the rate at which stats are recorded for <kind>, <op> takes a list of agents and a list of slots as
  arguments and produces a result, which will be stored indexed by the cycle number and then by <kind>

  some basic numbers, e.g., the number of objects of each kind at any update cycle are automatically recorded... (alivefreq ???)
  SHOULD WE ALLOW FOR THE RECORDING OF DEAD AGENTS?


Call the simulation with the parameter list as its only argument:

  simulate(params);

*/

;;; overwrites the same function in simulator.p
;;; adds a GUI contructor for a given type only if it is not present yet
;;; in which case the GUI constructor does not exist, it will be defined
;;; based on "basic_agentGUI" constructors will be stored for GUI and
;;; non-GUI agent types
define add_constructor_check(agtype);
    unless constructors(agtype) then
	;;; check if constructor exists, which it should if the GUI
	;;; class has been defined
	lvars GUIagtype = consword(agtype >< 'GUI'),
	     GUIcons = consword('new' >< agtype >< 'GUI');
	;;; check if the GUI class is defined for the agent, and if
	;;; it isn't, compile it on the fly
	;;; this needs to be taken out probably for the standalone
	;;; version because of limitations with POPC
 	if isundef(valof(GUIcons)) then
	    compile(stringin('define :class '>< GUIagtype >< '; is '
		    	     >< agtype >< ', basic_agentGUI; enddefine;'));
	endif;
	;;; now the value of GUIcons is defined, so store the new
	;;; constructor in the constructor property
    valof(GUIcons) ->> constructors(agtype) -> constructors(GUIagtype);
    endunless;
enddefine;

define startsimulation(params);
    ;;; setup graphics
    lvars win_spec = [^window_x ^window_y ^window_width ^window_height
                           ^window_frame ^window_title ^window_transform],
          world_win_spec = [^world_window_x ^world_window_y ^world_window_width ^world_window_height
                           ^world_window_frame ^world_window_title ^world_window_transform],
         win_frame, world_win_frame, win;

    rc_new_window_object(
            		 explode(win_spec) ->win_frame, newsim_picagent_window) -> sim_windowlist;
    fill(explode(win_frame), sim_picframe(sim_windowlist)) ->;

    ;;; Make the window mouse and key sensitive
    rc_mousepic(sim_windowlist);


	rc_new_window_object(
            		 explode(world_win_spec) ->world_win_frame, newsim_picagent_window) -> win;
    fill(explode(win_frame), sim_picframe(win)) ->;

    ;;; make window list
	conslist(sim_windowlist,win,2) -> sim_windowlist;

	;;; now draw the the boundaries of the simworld window in the worldview window...
	
    ;;; make it the current one
    sim_windowlist(2) -> rc_current_window_object;
    lvars fg = rc_foreground(rc_window);
    'black' -> rc_foreground(rc_window);
    rc_drawline(-window_lim_x,-window_lim_y,window_lim_x,-window_lim_y);
    rc_drawline(window_lim_x,-window_lim_y,window_lim_x,window_lim_y);
    rc_drawline(window_lim_x,window_lim_y,-window_lim_x,window_lim_y);
    rc_drawline(-window_lim_x,window_lim_y,-window_lim_x,-window_lim_y);
    fg -> rc_foreground(rc_window);


    /*  NOT QUITE CLEAR WHAT THIS DOES...
        if isrc_window_object(sim_panel) and rc_widget(sim_panel) then
            rc_redraw_panel(sim_panel)
        else
            lvars oldwin_obj = rc_current_window_object;
            ;;; ADD control for panel for the simulation...
            ;;; sim_control_panel(10,10,sim_alltracevars,2,'Control');
            oldwin_obj -> rc_current_window_object;
        endif;
        */
    simulate(params);
enddefine;

;;; simulate one step at a time in graphics mode
define startsimulationstep(params);
    dlocal sim_pause_demo = true;
    startsimulation(params);
enddefine;

;;; continues an interrupted simulation
define continue(runtime);
    dlocal sim_pause_demo = false;
    unless isundef(alive_agents) then
        ;;; Specify decimal places when printing numbers
        dlocal pop_pr_places = sim_pr_places;
        ;;; run it
        sim_scheduler_objects(alive_agents,runtime) -> alive_agents;
    endunless;
enddefine;

;;; continues an interrupted simulation in step mode
define continuestep(runtime);
    dlocal sim_pause_demo = true;
    continue(runtime);
enddefine;

;;; closes the simulation windows
define close_environment();
    rc_kill_window_object(sim_windowlist(1));
    rc_kill_window_object(sim_windowlist(2));
enddefine;

;;; computes some more stats and prints out who is left after a whole simulation, in HTML or ASCII
define printagentstats(agents,which,format);
    lvars a = 0, c = 0, d = 0, r = 0,
          entity, mat = [], coef = [],
	  agent_sten = newproperty([], 16, 0.0, "perm"),
	  agent_stwa = newproperty([], 16, 0.0, "perm"),
	  agent_pren = newproperty([], 16, 0.0, "perm"),
	  agent_prwa = newproperty([], 16, 0.0, "perm");

    for entity in agents do
        if isbasic_agent(entity) then
	    ;;; this may write the same information many times, that's ok, just redundant
	    startupenergy(entity) -> agent_sten(kind(entity));
	    startupwater(entity) -> agent_stwa(kind(entity));
	    startupenergy(entity) -> agent_pren(kind(entity));
	    startupwater(entity) -> agent_prwa(kind(entity));
            agent_alive(kind(entity)) + 1 -> agent_alive(kind(entity));
	    agent_age(kind(entity)) + age(entity) -> agent_age(kind(entity));
        endif;
    endfor;

    ;;; print statistics as plain text or as HTML (using tables)
    lvars ag;
    if format == "HTML" then
        printf('<table>\n');
        printf('<tr><td>PREY</td><td>total</td><td>alive</td><td>crashed</td>'
	    >< '<td>eaten</td><td>hunger</td><td>thirst</td><td>av.age</td>'
	    >< '<td>av.eng</td><td>av.wat</td><td>av.dis</td><td>av.off</td>'
	    >< '<td>eff</td><td>wef</td></tr>\n');
    elseif format == "ASCII" then
	printf('PREY        total alive crashed eaten hunger thirst av.age'
	    >< 'av.energy av.water av.dist av.offspr eef. wef.\n');
    endif;
    for ag in which do
	consword(ag >< 'GUI') -> ag;
        lvars tot = agent_eaten(ag) +  agent_crash(ag) +
             agent_hunger(ag)+ agent_thirst(ag) + agent_alive(ag);
	;;;npr('ag: ' >< ag >< ', tot = ' >< tot);
        if tot /= 0 then
	    ;;;npr('tot = ' >< tot);
	    lvars av_off = agent_offspring(ag)/tot,
		 av_dis = agent_distance(ag)/tot,
		 av_eng = agent_energy(ag)/tot,
 		 av_wat = agent_water(ag)/tot,
		 av_age = agent_age(ag)/tot,
		 ;;; get the energy and water expenditure per distance unit
		 ;;; travelled...
		 ;;; Need to exclude computation cost from these
		 ;;; calculations, hence the "- av_age" part.  In the
		 ;;; future, this will need to be more sophisticated.
		 av_eef = (agent_sten(ag) + av_eng - av_off * agent_pren(ag)
		    - av_age) / av_dis,
		 av_wef = (agent_stwa(ag) + av_wat - av_off * agent_prwa(ag)
		    - av_age) / av_dis;
	    if format == "HTML" then
		printf('<tr><td>%p:</td><td>%p</td><td>%p</td><td>%p</td>'
		>< '<td>%p</td><td>%p</td><td>%p</td><td>%p</td>'
		>< '<td>%p</td><td>%p</td><td>%p</td><td>%p</td>'
		>< '<td>%p</td><td>%p</td></tr>\n',
                [%ag,tot,agent_alive(ag),agent_crash(ag),agent_eaten(ag),
                agent_hunger(ag),agent_thirst(ag), av_age, av_eng, av_wat,
		av_dis, av_off, av_eef, av_wef%]);
	    elseif format == "ASCII" then
                printf('%p: %p  %p  %p  %p  %p  %p  %p  %p  %p  %p  %p  %p'
		>< '%p\n',
                [%ag,tot,agent_alive(ag),agent_crash(ag),agent_eaten(ag),
                agent_hunger(ag),agent_thirst(ag), av_age, av_eng, av_wat,
		av_dis, av_off, av_eef, av_wef%]);
	    endif;
	    ;;; save the averages in the respective properties, so they
	    ;;; won't have to be computed in the server
	    ;;; the original figures can be easily retrieved by multiplying
	    ;;; the values by total number of agents
	    av_off -> agent_offspring(ag);
	    av_dis -> agent_distance(ag);
	    av_eng -> agent_energy(ag);
 	    av_wat -> agent_water(ag);
	    av_age -> agent_age(ag);
	    av_eef -> agent_eef(ag);
	    av_wef -> agent_wef(ag);
        endif;
    endfor;
    if format == "HTML" then
	printf('</table>\n');
    endif;
enddefine;
